using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SimpleJSON;
using System.Linq;
using MVR.FileManagementSecure;
using static ClockwiseSilver.Helpers;

namespace ClockwiseSilver
{
	public class Expression
	{
		public string name = "";
		public DAZMorph morph;
		public float target = 1f;
		public float holdTime = 1f;
		public float speed = 1f;
		
		public Expression(string name, DAZMorph morph)
		{
			this.name = name;
			this.morph = morph;
			morph.morphValue = 0f;
		}
	}
	
	public class ExpressionToolPlus : MVRScript
	{
		private Atom her;
		private GenerateDAZMorphsControlUI morphControl;
		private JSONStorableBool isActive;		
		private JSONStorableFloat morphTime1JSON, morphSpeed1JSON, morphMin1JSON, morphMax1JSON;
		private JSONStorableFloat morphTime2JSON, morphSpeed2JSON, morphMin2JSON, morphMax2JSON;
		private JSONStorableFloat updateTimeJSON;
		
		private JSONStorableString sString1;
		private UIDynamicTextField sText1;
		private List<string> sequence1 = new List<string>();
		private JSONStorableStringChooser jchooser1;
		private UIDynamicPopup popup1;
		
		private JSONStorableString sString2;
		private UIDynamicTextField sText2;
		private List<string> sequence2 = new List<string>();
		private JSONStorableStringChooser jchooser2;
		private UIDynamicPopup popup2;
		
		private bool stopping = false;
		private List<Expression> expList1 = new List<Expression>();
		private List<Expression> expList2 = new List<Expression>();
		
		private string regionName = "exp";
		private bool initialized = false;
		private float updateTime = 0.166f;
		private float speedMod = 16.6f;
		
		//Credit to Physis for json save/load process
		//https://www.reddit.com/r/VAMscenes/comments/alhc6j/plugin_movement_helper/
		public static string saveExt = "exa";
		private string _lastBrowseDir;
		
		public override void Init() {
			try
			{
				her = containingAtom;
                JSONStorable geometry = her.GetStorableByID("geometry");
                DAZCharacterSelector character = geometry as DAZCharacterSelector;
                morphControl = character.morphsControlUI;
				List<string> choices = new List<string>();
				choices.Add("None");
			
				if (morphControl != null)
				{
					morphControl.GetMorphDisplayNames().ForEach((name) =>
					{
						DAZMorph morph = morphControl.GetMorphByDisplayName(name);
						if (morph.region.ToLower().Contains(regionName))
						{
							choices.Add(name);
						}
					});
				}
				
				isActive = new JSONStorableBool ("isActive", false);
				RegisterBool (isActive);
				var toggle = CreateToggle(isActive);
				toggle.label = "Active";
				
				JSONStorableAction ToggleActive = new JSONStorableAction("Toggle Expressions", () =>
                {
                    isActive.val = !isActive.val;
                });
                RegisterAction(ToggleActive);
				
				var heading = CreateTextField(new JSONStorableString("header0",
					"Requires at least 3 expressions \nin each section.\n" +
					"Avoid using the same expression\nin both sections.\n" +
					"Avoid using formula expressions\nthat drive other parameters."));
				heading.height = 200f;
				
				//-----------------UI------------------
				
				morphTime1JSON = new JSONStorableFloat("Change Exp Time1", 2.5f, 0.05f, 8.0f, false);
				RegisterFloat(morphTime1JSON);
				CreateSlider(morphTime1JSON, false);
				
				morphSpeed1JSON = new JSONStorableFloat("Morph Speed1", 1.0f, 0.1f, 4.0f, false);
				RegisterFloat(morphSpeed1JSON);
				CreateSlider(morphSpeed1JSON, false);
				
				morphMin1JSON = new JSONStorableFloat("Expressions Min1", 0.25f, 0.05f, 1.0f, false);
				RegisterFloat(morphMin1JSON);
				CreateSlider(morphMin1JSON, false);
				
				morphMax1JSON = new JSONStorableFloat("Expressions Max1", 0.55f, 0.05f, 1.0f, false);
				RegisterFloat(morphMax1JSON);
				CreateSlider(morphMax1JSON, false);
				
				morphTime2JSON = new JSONStorableFloat("Change Exp Time2", 3f, 0.05f, 8.0f, false);
				RegisterFloat(morphTime2JSON);
				CreateSlider(morphTime2JSON, false);
				
				morphSpeed2JSON = new JSONStorableFloat("Morph Speed2", 0.8f, 0.1f, 5.0f, false);
				RegisterFloat(morphSpeed2JSON);
				CreateSlider(morphSpeed2JSON, false);
				
				morphMin2JSON = new JSONStorableFloat("Expressions Min2", 0.25f, 0.05f, 1.0f, false);
				RegisterFloat(morphMin2JSON);
				CreateSlider(morphMin2JSON, false);
				
				morphMax2JSON = new JSONStorableFloat("Expressions Max2", 0.55f, 0.05f, 1.0f, false);
				RegisterFloat(morphMax2JSON);
				CreateSlider(morphMax2JSON, false);
				
				updateTimeJSON = new JSONStorableFloat("Update FrameRate", 60f, 1f, 120f, false);
				RegisterFloat(updateTimeJSON);
				CreateSlider(updateTimeJSON, false);
				
				var btn = CreateButton("Reset Expressions", false);
                btn.button.onClick.AddListener(() => { ResetExpressions(); });
				
				//------------------------------------Chooser 1------------------------------------
				sString1 = new JSONStorableString("Sequence1String", "");
				sString1.storeType = JSONStorableParam.StoreType.Full;
				RegisterString(sString1);
				sText1 = CreateTextField(sString1, true);
				sText1.height = 120;
				btn = CreateButton("Remove Last", true);
                btn.button.onClick.AddListener(() => { RemoveLastExpression(expList1, sString1); });
				btn.buttonColor = Color.gray;
				btn.textColor = Color.white;
				btn = CreateButton("Add Active", true);
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
                btn.button.onClick.AddListener(() => { AddActive(expList1, sString1); });					
				
				jchooser1 = new JSONStorableStringChooser("ExpChooser1", choices, "None", "Express", ExpressionsCallback1);                
				RegisterStringChooser(jchooser1);
				popup1 = CreateFilterablePopup(jchooser1, true);
				popup1.popupPanelHeight = 440f;
				
				sString2 = new JSONStorableString("Sequence2String", "");
				sString2.storeType = JSONStorableParam.StoreType.Full;
				RegisterString(sString2);
				sText2 = CreateTextField(sString2, true);
				sText2.height = 120;
				btn = CreateButton("Remove Last", true);
                btn.button.onClick.AddListener(() => { RemoveLastExpression(expList2, sString2); });
				btn.buttonColor = Color.gray;
				btn.textColor = Color.white;				
				btn = CreateButton("Add Active", true);
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
                btn.button.onClick.AddListener(() => { AddActive(expList2, sString2); });				
				
				jchooser2 = new JSONStorableStringChooser("ExpChooser2", choices, "None", "Express", ExpressionsCallback2);                
				RegisterStringChooser(jchooser2);
				popup2 = CreateFilterablePopup(jchooser2, true);
				popup2.popupPanelHeight = 440f;
				
				//------------------------Presets-------------------------
				CreateSpacer(true).height = 40;
				btn = CreateButton("Load Preset",true);
				btn.button.onClick.AddListener(() =>
				{
					SuperController.singleton.NormalizeMediaPath("Saves\\ExpToolPresets\\"); 
					SuperController.singleton.GetMediaPathDialog(LoadPreset, saveExt);
				});
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;

				btn = CreateButton("Save Preset", true);
				btn.button.onClick.AddListener(() =>
				{
					FileManagerSecure.CreateDirectory("Saves\\ExpToolPresets");
					SuperController.singleton.NormalizeMediaPath("Saves\\ExpToolPresets\\"); 
					SuperController.singleton.GetMediaPathDialog(SavePreset, saveExt);

					// Update the browser to be a Save browser
					uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
					browser.defaultPath =  "Saves\\ExpToolPresets";
					browser.SetTextEntry(true);
					browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), saveExt);
					browser.ActivateFileNameField();
				});
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}
		
		private void Start()
        {	
			StartCoroutine("StartRoutine");
        }
		
		private IEnumerator StartRoutine()
		{
			while (SuperController.singleton.isLoading)
			{
				yield return null;
			}
			
			ReloadExpressions();
			if (expList1.Count == 0) { Preset1(); }
			if (expList2.Count == 0) { Preset2(); }
			
			if (isActive.val && expList1.Count != 0 && expList2.Count != 0)
			{
				StartExpressions();
			}
			
			//Set Listeners
			isActive.toggle.onValueChanged.AddListener(checkVal =>
				{
					if (checkVal) { StartExpressions(); }
					else { StopExpressions(); }
				});
			
			updateTime = 1f / updateTimeJSON.val;
			speedMod = updateTime * 100;
			updateTimeJSON.slider.onValueChanged.AddListener((value) => { updateTime = 1f / value; speedMod = updateTime * 100; });	
			
			initialized = true;
		}
		
		private void AddActive(List<Expression> eList, JSONStorableString sString)
		{
			if (isActive.val) { isActive.val = false; }
			
			morphControl.GetMorphDisplayNames().ForEach((name) =>
			{
				DAZMorph m = morphControl.GetMorphByDisplayName(name);
				if (m != null && m.region.ToLower().Contains(regionName) && m.morphValue != 0f)
				{
					AddExpression(name, eList);
				}
			});
			
			SetExpressionString(eList, sString);
		}
		
		void SavePreset(string presetPath)
        {
            if (String.IsNullOrEmpty(presetPath)) { return; }
            _lastBrowseDir = presetPath.Substring(0, presetPath.LastIndexOfAny(new char[] { '/', '\\' })) + @"\";

            if ( !presetPath.ToLower().EndsWith(saveExt.ToLower()))
            {
                presetPath += "." + saveExt;
            }
            JSONClass saveJson = new JSONClass();
            saveJson["sequence1"] = new JSONArray();
            saveJson["sequence2"] = new JSONArray();
			saveJson["settings"] = new JSONArray();
            foreach (Expression e in expList1)
            {
				saveJson["sequence1"].Add(e.name);
            }
			foreach (Expression e in expList2)
            {
				saveJson["sequence2"].Add(e.name);
            }
			saveJson["settings"].Add(morphTime1JSON.val.ToString());
			saveJson["settings"].Add(morphSpeed1JSON.val.ToString());
			saveJson["settings"].Add(morphMin1JSON.val.ToString());
			saveJson["settings"].Add(morphMax1JSON.val.ToString());
			saveJson["settings"].Add(morphTime2JSON.val.ToString());
			saveJson["settings"].Add(morphSpeed2JSON.val.ToString());
			saveJson["settings"].Add(morphMin2JSON.val.ToString());
			saveJson["settings"].Add(morphMax2JSON.val.ToString());
			
            this.SaveJSON(saveJson, presetPath);
        }

        void LoadPreset(string presetPath)
        {
            if (String.IsNullOrEmpty(presetPath)) { return; }			
			if (isActive.val) { isActive.val = false; }
			
            _lastBrowseDir = presetPath.Substring(0, presetPath.LastIndexOfAny(new char[] { '/', '\\' })) + @"\";
            var jsonFile = this.LoadJSON(presetPath);

			ClearExpressions(expList1);
			foreach (JSONNode nd in jsonFile["sequence1"].AsArray)
            {
                AddExpression(nd.Value, expList1);
            }
			SetExpressionString(expList1, sString1);
			
			ClearExpressions(expList2);
			foreach (JSONNode nd in jsonFile["sequence2"].AsArray)
            {
                AddExpression(nd.Value, expList2);
            }
			SetExpressionString(expList2, sString2);
			
			if (jsonFile["settings"][0] != null) { morphTime1JSON.val = jsonFile["settings"][0].AsFloat; }
			if (jsonFile["settings"][1] != null) { morphSpeed1JSON.val = jsonFile["settings"][1].AsFloat; }
			if (jsonFile["settings"][2] != null) { morphMin1JSON.val = jsonFile["settings"][2].AsFloat; }
			if (jsonFile["settings"][3] != null) { morphMax1JSON.val = jsonFile["settings"][3].AsFloat; }
			if (jsonFile["settings"][4] != null) { morphTime2JSON.val = jsonFile["settings"][4].AsFloat; }
			if (jsonFile["settings"][5] != null) { morphSpeed2JSON.val = jsonFile["settings"][5].AsFloat; }
			if (jsonFile["settings"][6] != null) { morphMin2JSON.val = jsonFile["settings"][6].AsFloat; }
			if (jsonFile["settings"][7] != null) { morphMax2JSON.val = jsonFile["settings"][7].AsFloat; }
        }
		
		private void ReloadExpressions()
		{			
			ClearExpressions(expList1);
			string[] splitString = Regex.Split(sString1.val, ",,");			
			foreach (string str in splitString)
			{
				if (str != "") { AddExpression(str, expList1); }
			}
			
			ClearExpressions(expList2);
			splitString = Regex.Split(sString2.val, ",,");			
			foreach (string str in splitString)
			{
				if (str != "") { AddExpression(str, expList2); }
			}
		}
		
		private void RemoveDuplicates()
		{
			List<Expression> rList = new List<Expression>();
			foreach (Expression e in expList1)
			{
				if (ContainsExpression(e.name, expList2))
				{
					rList.Add(e);
				}
			}
			foreach (Expression e in rList)
			{
				expList1.Remove(e);
			}
		}
		
		private void ClearExpressions(List<Expression> eList)
		{
			eList.Clear();
		}
		
		private void AddExpression(string name, List<Expression> eList)
		{
			if (String.IsNullOrEmpty(name) || ContainsExpression(name, eList)) { return; }
			DAZMorph m = morphControl.GetMorphByDisplayName(name);					
			if (m != null)
			{
				Expression e = new Expression(name, m);
				eList.Add(e);
			}
			RemoveDuplicates();
		}
		
		private bool ContainsExpression(string name, List<Expression> eList)
		{
			foreach (Expression e in eList)
			{
				if (e.name == name) { return true; }
			}
			return false;
		}
		
		private void SetExpressionString(List<Expression> eList, JSONStorableString sString)
		{
			sString.val = "";
			for (int i = 0; i < eList.Count; i++)
			{
				if (i > 0) { sString.val += ",,"; }
				sString.val += eList[i].name;
			}
		}
		
		private void ResetExpressions()
		{
			morphControl.GetMorphDisplayNames().ForEach((name) =>
                {
					if (!String.IsNullOrEmpty(name))
					{
						DAZMorph m = morphControl.GetMorphByDisplayName(name);
						if (m != null && m.region.ToLower().Contains(regionName))
						{
							m.morphValue = 0.0f;
						}
					}
                });
		}
		
		private void RemoveLastExpression(List<Expression> eList, JSONStorableString sString)
		{
			if (isActive.val) { isActive.val = false; }			
			eList.RemoveAt(eList.Count - 1);
			SetExpressionString(eList, sString);
		}
		
		private void Preset1()
		{
			if (isActive.val) { isActive.val = false; }
			sString1.val = "";
			ClearExpressions(expList1);
			AddExpression("Smile Full Face", expList1);
			AddExpression("Smile Open Full Face", expList1);
			AddExpression("Desire", expList1);
			AddExpression("Snarl Left", expList1);
			AddExpression("Flirting", expList1);			
			SetExpressionString(expList1, sString1);
		}
		
		private void Preset2()
		{
			if (isActive.val) { isActive.val = false; }
			sString2.val = "";
			ClearExpressions(expList2);
			AddExpression("Concentrate", expList2);
			AddExpression("Glare", expList2);
			AddExpression("Shock", expList2);
			AddExpression("Surprise", expList2);
			SetExpressionString(expList2, sString2);
		}
		
		private void ExpressionsCallback1(string s)
		{
			if (!initialized || s == "None" || String.IsNullOrEmpty(s)) { return; }
			if (isActive.val) { isActive.val = false; }			
			AddExpression(s, expList1);			
			SetExpressionString(expList1, sString1);
			jchooser1.val = "None";
		}
		
		private void ExpressionsCallback2(string s)
		{
			if (!initialized || s == "None" || String.IsNullOrEmpty(s)) { return; }
			if (isActive.val) { isActive.val = false; }			
			AddExpression(s, expList2);			
			SetExpressionString(expList2, sString2);
			jchooser2.val = "None";
		}
		
		private Expression GetTargetExpression(Expression eCurrent, List<Expression> eList, bool isOne=true)
		{
			UnityEngine.Random.InitState((int)(Time.time + UnityEngine.Random.Range(1, 1000)));
			int randomIndex = (int)UnityEngine.Random.Range(0, eList.Count);
			if (randomIndex == eList.Count) { randomIndex = 0; }
			Expression e = eList[randomIndex];
			if (eCurrent != null && e == eCurrent)
			{
				randomIndex++;
				if (randomIndex >= eList.Count) { randomIndex = 0; }
				e = eList[randomIndex];
			}
			
			if (isOne)
			{
				e.holdTime = morphTime1JSON.val * UnityEngine.Random.Range(0.75f, 1.25f);			
				e.target = UnityEngine.Random.Range(morphMin1JSON.val, morphMax1JSON.val);
				e.speed = morphSpeed1JSON.val;
			}
			else
			{
				e.holdTime = morphTime2JSON.val * UnityEngine.Random.Range(0.75f, 1.25f);			
				e.target = UnityEngine.Random.Range(morphMin2JSON.val, morphMax2JSON.val);
				e.speed = morphSpeed2JSON.val;
			}
			
			return e;
		}
		
		public void StartExpressions()
		{			
			ResetExpressions();
			stopping = false;
			StartCoroutine(MorphIt(GetTargetExpression(null, expList1, true), expList1, true));
			StartCoroutine(MorphIt(GetTargetExpression(null, expList2, false), expList2, false));
		}
		
		public void StopExpressions()
		{
			stopping = true;
		}
		
		private IEnumerator MorphIt(Expression e, List<Expression> eList, bool isOne)
		{		
			if (isOne) { yield return new WaitForSeconds(updateTime * 0.47f); }
			float val = 0;
			while (!stopping && val < 1f)
			{
				val += updateTime * e.speed;
				e.morph.morphValue = Mathf.Clamp(Helpers.QuadraticInOut(val) * e.target, 0f, 1f);
				yield return new WaitForSeconds(updateTime);
			}
			
			val = 0;
			while (!stopping && val < e.holdTime)
			{
				val += updateTime;
				yield return new WaitForSeconds(updateTime);
			}
			
			//Start dialing up the new morph WHILE dialing down the current
			if (!stopping)
			{
				StartCoroutine(MorphIt(GetTargetExpression(e, eList, isOne), eList, isOne));
			}
			
			val = 1;			
			while (!stopping && val > 0f)
			{
				val -= updateTime * e.speed;
				e.morph.morphValue = Mathf.Clamp(Helpers.QuadraticInOut(val) * e.target, 0f, 1f);
				yield return new WaitForSeconds(updateTime);
			}
			
			e.morph.morphValue = 0f;
		}

		private void OnDisable()
		{
			StopExpressions();
			ResetExpressions();
		}
		
		private void OnDestroy()
		{
			StopExpressions();
			ResetExpressions();
		}
	}
}