﻿using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;

namespace ClockwiseSilver {
	public static class Helpers
	{
		public static List<string> GetFemaleChoices(JSONStorableStringChooser chooser=null)
        {
			List<string> choices = SuperController.singleton.GetAtoms().Where(a => a.category == "People" && a.GetComponentInChildren<DAZCharacter>().name.StartsWith("female")).Select(a => a.name).ToList();
			if (chooser != null)
			{
				chooser.choices = choices;
				if (chooser.val == null && choices.Count > 0)
				{
					chooser.val = choices[0];
				}
			}
			return choices;
        }
		
		public static List<string> GetMaleChoices(JSONStorableStringChooser chooser=null)
        {
			List<string> choices = SuperController.singleton.GetAtoms().Where(a => a.category == "People" && !a.GetComponentInChildren<DAZCharacter>().name.Contains("female")).Select(a => a.name).ToList();
			if (chooser != null)
			{
				chooser.choices = choices;
				if (chooser.val == null && choices.Count > 0)
				{
					chooser.val = choices[0];
				}
			}
			return choices;
        }
		
		public static List<string> GetMaleAndToyChoices(JSONStorableStringChooser chooser=null)
        {
            List<string> choices = new List<string>();
			List<string> mChoices = SuperController.singleton.GetAtoms().Where(a => a.category == "People" && !a.GetComponentInChildren<DAZCharacter>().name.Contains("female")).Select(a => a.name).ToList();
            List<string> dChoices = SuperController.singleton.GetAtoms().Where(atom => atom.name.StartsWith("Dildo")).Select(atom => atom.name).ToList();
			List<string> cChoices = SuperController.singleton.GetAtoms().Where(atom => atom.name.StartsWith("Cock")).Select(atom => atom.name).ToList();			
			
			for (int i = 0; i < mChoices.Count; i++)
			{
				choices.Add(mChoices[i]);
			}
			for (int i = 0; i < dChoices.Count; i++)
			{
				choices.Add(dChoices[i]);
			}
			for (int i = 0; i < cChoices.Count; i++)
			{
				choices.Add(cChoices[i]);
			}
			
			if (chooser != null)
			{
				chooser.choices = choices;
				if (chooser.val == null && choices.Count > 0)
				{
					chooser.val = choices[0];
				}
			}
			return choices;
        }
		
		public static List<JSONStorable> FindPersonPluginInstances(string fullClassName)
		{
			List<Atom> atoms = SuperController.singleton.GetAtoms().Where(a => a.category == "People").ToList();
			List<JSONStorable> storables = new List<JSONStorable>();
			foreach (Atom a in atoms)
			{
				List<string> names = a.GetStorableIDs();
				string pluginName = names.Find(s => s.StartsWith("plugin#") && s.Contains(fullClassName));
				if (pluginName != null && pluginName != "")
				{
					storables.Add(a.GetStorableByID(pluginName));
				}
			}
			return storables;
		}
		
		public static string[] CommonFreeControllers()
		{
			return new string[] { "lHandControl", "rHandControl", "lFootControl", "rFootControl", "lKneeControl", "rKneeControl",
				"lArmControl", "rArmControl", "lElbowControl", "rElbowControl",	"lThighControl", "rThighControl", "pelvisControl",
				"hipControl", "chestControl", "neckControl", "headControl", "abdomenControl", "abdomen2Control"};
		}
		
		public static string[] CommonRigidbodies()
		{
			return new string[] { "lHand", "rHand", "lShldr", "rShldr", "lForeArm", "rForeArm", "lFoot", "rFoot",
				"lShin", "rShin", "lThigh", "rThigh", "pelvis", "hip", "chest", "neck", "head", "abdomen", "abdomen2"};
		}
		
		public static string RigidbodyToFreeController(string rigidbodyName)
		{
			return rigidbodyName.Replace("Shldr", "Arm").Replace("ForeArm", "Elbow").Replace("Shin", "Knee") + "Control";
		}
		
		public static string FreeControllerToRigidbody(string freeControllerName)
		{
			return freeControllerName.Replace("Arm", "Shldr").Replace("Elbow", "ForeArm").Replace("Control", "");
		}
		
		public static bool GetStoredNode(MVRScript script, string storableName, out JSONNode node)
		{
			foreach (JSONNode a in SuperController.singleton.loadJson["atoms"].AsArray)
			{
				if (a["id"].Value == script.containingAtom.name)
				{
					foreach (JSONNode n in a["storables"].AsArray)
					{
						if (n["id"].Value == script.storeId)
						{
							if (n[storableName].Value != "")
							{
								node = n[storableName];
								return true;
							}
						}
					}
				}
			}	
			node = null;
			return false;
		}
		
		public static float GetStoredFloat(MVRScript script, string storableName, float fallbackValue=0f)
		{
			JSONNode node;
			if (GetStoredNode(script, storableName, out node))
			{
				return node.AsFloat;
			}
			return fallbackValue;
		}
		
		public static bool GetStoredBool(MVRScript script, string storableName, bool fallbackValue=false)
		{
			JSONNode node;
			if (GetStoredNode(script, storableName, out node))
			{
				return node.AsBool;
			}
			return fallbackValue;
		}
		
		public static string GetStoredString(MVRScript script, string storableName, string fallbackValue="")
		{
			JSONNode node;
			if (GetStoredNode(script, storableName, out node))
			{
				return node.Value;
			}
			return fallbackValue;
		}
		
		public static void RestoreStoredFloats(MVRScript script, List<JSONStorableFloat> sliders)
		{
			foreach (JSONStorableFloat slider in sliders)
			{
				slider.val = GetStoredFloat(script, slider.name, slider.defaultVal);
			}
		}
		
		public static void RestoreStoredBools(MVRScript script, List<JSONStorableBool> toggles)
		{
			foreach (JSONStorableBool toggle in toggles)
			{
				toggle.val = GetStoredBool(script, toggle.name, toggle.defaultVal);
			}
		}
		
		public static void RestoreStoredStrings(MVRScript script, List<JSONStorableString> strs)
		{
			foreach (JSONStorableString str in strs)
			{
				str.val = GetStoredString(script, str.name, "");
			}
		}
		
		public static float Rollover(float x)
		{
			if (x > 360f) { return x - 360f; }
			else if (x < 0f) { return 360f - x; }
			return x;
		}
		
		//https://referencesource.microsoft.com/#System.Numerics/System/Numerics/Quaternion.cs
		public static Quaternion NormalizeQuaternion(Quaternion q)
        {
            Quaternion ans; 
            float ls = q.x * q.x + q.y * q.y + q.z * q.z + q.w * q.w; 
            float invNorm = 1.0f / (float)Math.Sqrt((double)ls); 
            ans.x = q.x * invNorm;
            ans.y = q.y * invNorm;
            ans.z = q.z * invNorm;
            ans.w = q.w * invNorm; 
            return ans;
        }
		
		//----------------------------------------------------------Interpolation
		//Catmull-Rom implementaiton - Credit https://www.habrador.com/tutorials/interpolation/1-catmull-rom-splines/
		public static Vector3 GetCatmullRomPosition(float t, Vector3 p0, Vector3 p1, Vector3 p2, Vector3 p3)
		{
			//The coefficients of the cubic polynomial (except the 0.5f * which I added later for performance)
			Vector3 a = 2f * p1;
			Vector3 b = p2 - p0;
			Vector3 c = 2f * p0 - 5f * p1 + 4f * p2 - p3;
			Vector3 d = -p0 + 3f * p1 - 3f * p2 + p3;

			//The cubic polynomial: a + b * t + c * t^2 + d * t^3
			return 0.5f * (a + (b * t) + (c * t * t) + (d * t * t * t));
		}
		
		//----------------------------------------------------------Easings
		public static float QuadraticInOut(float x)
		{
			if ((x *= 2f) < 1f) return 0.5f * x * x;
            return -0.5f * ((x -= 1f) * (x - 2f) - 1f);
		}
		
		public static float QuartOut(float x)
		{
			return 1 - Mathf.Pow(1 - x, 4);
		}
		
		//----------------------------------------------------------UI
		
		public static JSONStorableString SetupHeader(MVRScript script, string label, string text, bool right, float height=10f)
		{
			JSONStorableString header = new JSONStorableString(label, text);
			script.CreateTextField(header, right).height = height;
			return header;
		}
		
		//----------------------------------------------------------Materials
		
		
		//----------------------------------------------------------MacGruber Utils
		//Credit MacGruber 2020
		public static JSONStorable FindPlugin(MVRScript self, string fullClassName)
		{
			List<string> names = self.containingAtom.GetStorableIDs();
			//foreach (string str in names) { SuperController.LogMessage(str); }
			string pluginName = names.Find(s => s.StartsWith("plugin#") && s.Contains(fullClassName));
			if (pluginName != null && pluginName != "")
			{
				return self.containingAtom.GetStorableByID(pluginName);
			}
			else
			{
				return null;
			}
		}
		// Create VaM-UI Toggle button
		public static JSONStorableBool SetupToggle(MVRScript script, string label, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(label, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateToggle(storable, rightSide);
			script.RegisterBool(storable);
			return storable;
		}
		
		// Create VaM-UI Float slider
		public static JSONStorableFloat SetupSlider(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateSlider(storable, rightSide);
			script.RegisterFloat(storable);
			return storable;
		}
		
		// Create VaM-UI ColorPicker
		public static JSONStorableColor SetupColor(MVRScript script, string label, Color color, bool rightSide)
		{
			HSVColor hsvColor = HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
			JSONStorableColor storable = new JSONStorableColor(label, hsvColor);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateColorPicker(storable, rightSide);
			script.RegisterColor(storable);
			return storable;
		}
		
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
	}
}